from ._base import WorldObject


class Group(WorldObject):
    """A Group is almost identical to a WorldObject. Its purpose is
    to make working with groups of objects syntactically clearer.
    """
